/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.communication.jmx;

import java.util.NoSuchElementException;
import java.util.logging.Logger;

import javax.management.MBeanServer;

import org.apache.commons.pool2.KeyedObjectPool;
import org.apache.commons.pool2.impl.GenericKeyedObjectPool;
import org.apache.commons.pool2.impl.GenericKeyedObjectPoolConfig;
import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.objectweb.fractal.jmx.agent.AdminAttributes;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;

/**
 * This class is a manager of JMX server. 
 * <p>
 * It is used to get a connection to a JMX Server on a distant container.
 * </p>
 * 
 * @author ddesjardins - EBM WebSourcing
 * @author Christophe Hamerling - EBM WebSourcing
 * @author Roland Naudin - EBM WebSourcing
 */
@Component(provides = @Interface(name = "service", signature = JMXService.class))
public class JMXServiceImpl implements JMXService {

    /**
     * Logger
     */
    protected final LoggingUtil log = new LoggingUtil(Logger.getLogger(JMXService.COMPONENT_LOGGER_NAME));

    /**
     * Admin Attributes Fractal component
     */
    @Requires(name = "adminAtt")
    protected AdminAttributes adminAttributes;

    /**
     * Configuration service fractal component
     */
    @Requires(name = "configuration")
    protected ConfigurationService configurationService;

    /**
     * Map of PEtALS JMX connections
     */
    private KeyedObjectPool<String, JMXClient> petalsJmxClients;

    /**
     * The local container name
     */
    private String containerName;

    /**
     * {@inheritDoc}
     */
    public JMXClient getJMXClient(final String containerName) throws PetalsException {
        this.log.start("Container Name : " + containerName);

        try {
            return this.petalsJmxClients.borrowObject(containerName);
        } catch (final NoSuchElementException e) {
            throw new PetalsException(e);
        } catch (final IllegalStateException e) {
            throw new PetalsException(e);
        } catch (final Exception e) {
            throw new PetalsException(e);
        } finally {
            this.log.end();
        }
    }

    @Override
    public void returnJMXClient(final String containerName, final JMXClient petalsJmxClient) throws PetalsException {
        try {
            this.petalsJmxClients.returnObject(containerName, petalsJmxClient);
        } catch (final Exception e) {
            throw new PetalsException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    public MBeanServer getLocalJMXServer() {
        return this.adminAttributes.getRawMBeanServer();
    }

    @Lifecycle(step = Step.START)
    public void start() throws DuplicatedServiceException, MissingServiceException {
        this.log.start();

        this.containerName = this.configurationService.getContainerConfiguration().getName();

        // Initialize the Petals JMX API implementation
        PetalsJmxApiFactory.getInstance();

        final GenericKeyedObjectPoolConfig gopc = new GenericKeyedObjectPoolConfig();
        gopc.setBlockWhenExhausted(false);
        gopc.setTestOnBorrow(true);
        gopc.setTimeBetweenEvictionRunsMillis(600000l);
        this.petalsJmxClients = new GenericKeyedObjectPool<>(
                // object factory
                new JMXClientObjectFactory(this.containerName, this.getLocalJMXServer(), this.configurationService),
                gopc);

        this.log.end();
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        try {
            this.petalsJmxClients.close();
        } catch (final Exception e) {
            this.log.warning("An error occurs closing the JMX client pool.", e);
        }

        PetalsJmxApiFactory.close();

        this.log.end();
    }

}
